"""
Creates common preprocessing QC plots for an AnnData object
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
from anndata import AnnData
from typing import Optional, Tuple

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def qc_plotting(
        adata: AnnData,
        counts_thresholds: Tuple[int, int] = [1000, 30000],
        genes_thresholds: Tuple[int, int] = [100, 5000],
        mt_threshold: int = 10,
        show_thresholds: bool = True,
        batch_key: Optional[str] = None,
        show_legend: bool = True,
        figsize: Tuple[int, int] = (9, 3),
        dpi: int = 300,
        save_path: Optional[str] = None,
):
    """
    Creates common preprocessing QC plots for an AnnData object
    Parameters
    ----------
    adata
        The AnnData containing data to plot
    counts_thresholds
        The lower and upper thresholds to be used on total counts when filtering cells. Default (1000, 30000)
    genes_thresholds
        The lower and upper thresholds to be used on number of genes when filtering cells . Default (100, 5000)
    mt_threshold
        The threshold to be used on % mito reads when filtering cells. Default 10
    show_thresholds
        Whether to show the thresholds as dashed lines on each plot. Default True
    batch_key
        A column name in 'adata.obs' that annotates different batches of data for separate plotting. If 'None' treats
        all cells as coming from the same batch. Default None
    show_legend
        Whether to show the 'batch_key" labels as a legend. Default True
    figsize
        The size of the figure. Default (9, 3)
    dpi
        The resolution of the figure to save. Default 300
    save_path
        The path to save the figure to (/path/to/dir/filename). Default None

    Raises
    -------
    ValueError
        If batch_key is not a valid key in 'adata.obs.columns'
    """
    # Setup plots
    sns.set_style("ticks")
    sns.set_context("paper")
    fig = plt.figure(figsize=figsize)
    if batch_key is None:
        adata.obs["qc_plot_batch"] = "None"
        batch_key = "qc_plot_batch"
        show_legend = False
    else:
        if batch_key not in adata.obs.columns:
            raise ValueError(f"{batch_key} is not a valid column in 'adata.obs'")

    # Plot 1: Histogram of % mito counts
    plt.subplot(1, 3, 1)
    ax = sns.histplot(
        x=adata.obs.pct_counts_mt,
        hue=adata.obs[batch_key],
    )
    if not show_legend:
        ax.get_legend().remove()
    if show_thresholds:
        plt.axvline(
            x=mt_threshold,
            ymin=0,
            ymax=1,
            color="black",
            linestyle="--",
        )

    # Plot 2: total counts vs num genes
    plt.subplot(1, 3, 2)
    ax = sns.scatterplot(
        x=adata.obs.total_counts,
        y=adata.obs.n_genes_by_counts,
        s=2,
        hue=adata.obs[batch_key],
        linewidth=0,
    )
    if not show_legend:
        ax.get_legend().remove()
    if show_thresholds:
        for threshold in counts_thresholds:
            plt.axvline(
                x=threshold,
                ymin=0,
                ymax=1,
                color="black",
                linestyle="--",
            )
        for threshold in genes_thresholds:
            plt.axhline(
                y=threshold,
                xmin=0,
                xmax=1,
                color="black",
                linestyle="--",
            )

    # Plot 3: Rank ordered total counts
    plt.subplot(1, 3, 3)
    cell_data = pd.DataFrame()
    for category in adata.obs[batch_key].unique():
        category_data = adata[adata.obs[batch_key] == category].obs.copy()
        category_data["rank"] = category_data.total_counts.rank(method="first", ascending=False,)
        cell_data = pd.concat([cell_data, category_data])
    cell_data = cell_data.sort_values(by="rank")
    ax = sns.lineplot(
        x=cell_data["rank"],
        y=cell_data.total_counts,
        hue=cell_data[batch_key],
        hue_order=adata.obs[batch_key].unique(),
    )
    if not show_legend:
        ax.get_legend().remove()
    ax.set(yscale="log")
    ax.set(xscale="log")
    if show_thresholds:
        for threshold in counts_thresholds:
            plt.axhline(
                y=threshold,
                xmin=0,
                xmax=1,
                color="black",
                linestyle="--",
            )

    # Figure level adjustments
    fig.tight_layout()
    if save_path is not None:
        if "/" not in save_path:
            save_path = f"./{save_path}"
        check_path(save_path.rsplit("/", 1)[0])
        logger.info(f"Saving figure to {save_path}")
        plt.savefig(save_path, dpi=dpi, facecolor="white",)
    plt.show()



"""
Creates a pseudobulked cell type gene signature from scRNAseq data
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
from anndata import AnnData
from typing import Optional
import pandas as pd
import numpy as np

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def create_cell_type_signature(
        adata: AnnData,
        save_path: Optional[str] = None,
        cell_type_labels: str = "cell_type",
        data_loc: str = "raw",
) -> pd.DataFrame:
    """
    Creates a pseudobulked cell type gene signature from scRNAseq data

    Parameters
    ----------
        adata
            The AnnData object containing the cell type expression data, with cell labels as a key in 'adata.obs' and
            library size corrected and log-normalized counts in adata.X, adata.raw, or a layer in adata.layers.
        save_path
            The path to save a csv containing the cell type gene signatures to
        cell_type_labels
            The column name in 'adata.obs' containing the cell type labels for each cell. Default is 'cell_type'
        data_loc
            The location of the library size corrected and log-normalized gene expression data in 'adata'. Default is
            'raw' but can also be 'X' or a valid key from 'adata.layers'

    Returns
    -------
    A DataFrame containing the psuedobulked gene signatures for each cell type in adata.obs[cell_type_labels] with
    columns containing data each cell type and rows containing the average expression for each gene within that call
    type.

    Raises
    -------
    ValueError
        If 'cell_type_labels' is not a valid column in 'adata.obs'
    """
    if cell_type_labels not in adata.obs:
        raise ValueError(f"{cell_type_labels} not a valid column in 'adata.obs'")
    if data_loc == "raw":
        signatures = pd.DataFrame(index=adata.raw.var_names)
        raw_adata = adata.raw.to_adata()
        raw_adata.obs[cell_type_labels] = adata.obs[cell_type_labels]
    else:
        signatures = pd.DataFrame(index=adata.var_names)
    for cell_type in adata.obs[cell_type_labels].unique():
        if data_loc == "X":
            signatures[cell_type] = adata[adata.obs[cell_type_labels] == cell_type].X.mean(axis=0)
        elif data_loc == "raw":
            signatures[cell_type] = np.array(raw_adata[raw_adata.obs[cell_type_labels] == cell_type].X.todense()).mean(axis=0)
        else:
            if data_loc in adata.layers:
                signatures[cell_type] = adata[adata.obs[cell_type_labels] == cell_type].layers[data_loc].mean(axis=0)
            else:
                raise ValueError(f"{data_loc} not 'X', 'raw', or a valid layer in 'adata.layers'")
    if save_path is not None:
        check_path(save_path.rsplit("/")[0])
        logger.info(f"Saving signature DataFrame to {save_path}")
        signatures.to_csv(save_path)
    return signatures



"""
Filters out doublets using scrublet
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
from anndata import AnnData
from typing import Optional
import scrublet as scr
import pandas as pd

# scrnatools package imports
from .._configs import configs
from .._utils import debug

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def scrublet(
        adata: AnnData,
        raw_counts_layer: str,
        doublet_threshold: float = 0.2,
        batch_key: Optional[str] = None,
) -> AnnData:
    """
    Filters out doublets using scrublet

    Parameters
    ----------
    adata
        The AnnData to process
    doublet_threshold
        The doublet score threshold to call doublets/singlets on. Default 0.2
    batch_key
        A column in 'adata.obs' annotating different batches of data. If 'None' treats all cells as coming from the same
        batch. Default None
    raw_counts_layer
        A key in 'adata.layers' pointing to the raw counts data

    Returns
    -------
    The AnnData provided with doublets filtered out inplace based on doublet_threshold
    """
    # Setup
    scrublet_predictions = pd.DataFrame()
    if batch_key is None:
        adata.obs["scrublet_batch_key"] = "None"
        batch_key = "scrublet_batch_key"
    # Calculate doublet scores for each batch indepenently
    for key in adata.obs[batch_key].unique():
        subset_data = adata[adata.obs[batch_key] == key]
        counts_matrix = subset_data.layers[raw_counts_layer]
        scrub = scr.Scrublet(counts_matrix, expected_doublet_rate=0.06)
        doublet_scores, predicted_doublets = scrub.scrub_doublets(
            min_counts=2,
            min_cells=3,
            min_gene_variability_pctl=85,
            n_prin_comps=30
        )
        scrub.call_doublets(threshold=doublet_threshold)
        scrub.plot_histogram()
        scrublet_prediction = pd.DataFrame({"scrublet_score": doublet_scores}, index=subset_data.obs.index)
        scrublet_predictions = pd.concat([scrublet_predictions, scrublet_prediction])
    # Add scrublet scores to adata.obs and filter out cells with a score > doublet_threshold
    adata.obs["scrublet_score"] = scrublet_predictions.scrublet_score
    adata.obs["scrublet_called_doublet"] = adata.obs.scrublet_score > doublet_threshold
    num_doublets = adata.obs.scrublet_called_doublet.sum()
    pct_doublets = round(num_doublets / len(adata) * 100, 3)
    logger.info(f"{pct_doublets}% of cells classified as doublets ({num_doublets} cells)")
    return adata



"""
Plotting init script
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
from ._qc_plotting import qc_plotting
from ._gene_density_plot import gene_density_plot
from ._gene_embedding import gene_embedding
from ._isoform_qc_plot import isoform_qc_plot
from ._gene_heatmap import gene_heatmap
from ._gene_violinplot import gene_violinplot


"""
Calculates a differential density ratio of cells grouped by condition/control
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external package imports
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
from statsmodels.nonparametric.kernel_density import KDEMultivariateConditional,EstimatorSettings
from anndata import AnnData
from typing import Optional, Tuple

# scrnatools package imports
from .._configs import configs
from .._utils import debug

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def log_density_ratio(
        adata: AnnData,
        group_by: str,
        groups: Tuple[str, str],
        latent_rep: str = "X_umap",
        key_added: str = "log_density_ratio",
        est_settings: Optional = None,
):
    """
    Calculates a differential density ratio of cells grouped by condition/control
    Parameters
    ----------
    adata
        The AnnData containing labeled cell data
    group_by
        The column in 'adata.obs' containing condition and control labels
    groups
        The labels in 'adata.obs[group_by]' that distinguish condition and control cells, i.e. ["WT", "KO"]
    latent_rep
        The latent representation to calculate differential density across
    key_added
        The column in 'adata.obs' that will the density ratios will be added to
    est_settings

    Returns
    -------
        The original adata modified to contain the differential density ratios in a new column in 'adata.obs'

    Raises
    -------
    ValueError
        If 'group_by' is not a valid column name in 'adata.obs'
    ValueError
        If 'groups' are not valid labels in 'adata.obs[group_by]'
    """
    if group_by not in adata.obs:
        raise ValueError(f"{group_by} is not a valid column key in 'adata.obs'")
    for group in groups:
        if group not in adata.obs[group_by].unique():
            raise ValueError(
                f"{group} is not a valid category of 'adata.obs.{group_by}' ({adata.obs[group_by].unique()})"
            )
    density = KDEMultivariateConditional(
        endog=adata.obsm[latent_rep],
        exog=np.array(adata.obs[group_by].values == groups[1]),
        dep_type="cc",
        indep_type="u",
        bw="normal_reference",
        defaults=est_settings
    )
    logger.info(f"Calculating log density ratio of {groups[1]}/{groups[0]}")
    z1 = density.pdf(adata.obsm[latent_rep], 0 + np.zeros((adata.obsm[latent_rep].shape[0], 1)))
    z2 = density.pdf(adata.obsm[latent_rep], 1 + np.zeros((adata.obsm[latent_rep].shape[0], 1)))
    adata.obs[key_added] = np.log(z2/z1)  # group2 vs group1
    return adata



"""
Plots the density of expression of genes on an embedding
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
from anndata import AnnData
from typing import List, Optional, Tuple
import seaborn as sns
import matplotlib.pyplot as plt
from math import ceil
from scipy.sparse import issparse
from statsmodels.nonparametric.kernel_density import KDEMultivariateConditional, EstimatorSettings
from sklearn.preprocessing import StandardScaler
import numpy as np

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def gene_density_plot(
        adata: AnnData,
        gene_list: List[str],
        data_loc: str = "X",
        thresh: int = 1,
        latent_rep: str = "X_umap",
        est_settings=None,
        cmap: str = "magma",
        s: Optional[int] = None,
        ncols: int = 3,
        figsize: Tuple[int] = (3, 3),
        title: Optional[str] = None,
        save_path: Optional[str] = None,
        dpi: int = 300,
):
    """
    Plots the density of expression of genes on an embedding
    Parameters
    ----------
    adata
        The AnnData containing the gene expression and cell data
    gene_list
        A list of genes to plot
    data_loc
        The location of the expression data to use for density calculations, can be a layer in 'adata.layers' or 'X'
        to use the data stored in adata.X. Default 'X'
    thresh
    latent_rep
        The 2D representation to plot gene expression for each cell on. Default 'X_umap'
    est_settings
    cmap
        The pyplot colormap to use for plotting. Default 'magma'
    s
        The size of data points to plot. Default is to automatically calculate (None)
    ncols
        The number of columns in the figure
    figsize
        A tuple of the dimensions of each subplot for a single gene. Default (3,3)
    title
        The title of the figure. Default None
    save_path
        The path to save the figure. Default None.
    dpi
        The resolution of the saved image. Default 300

    Raises
    -------
    ValueError
        If the 'data_loc' provided is not 'X' or a valid layer in 'adata.layers'
    """

    # set up figure
    if len(gene_list) < ncols:
        ncols = len(gene_list)
    nrows = ceil((len(gene_list)) / ncols)
    fig = plt.figure(figsize=(ncols * figsize[0], nrows * figsize[1]))
    sns.set_theme(context="paper", style="white", )

    # Iterate over genes to plot
    for index, gene in enumerate(gene_list):
        if gene not in adata.var_names:
            logger.info(f"{gene} not in 'adata.var_names', skipping")
        else:
            temp_data = adata[:, gene]
            if data_loc == "X":
                if issparse(temp_data.X):
                    x = temp_data.X.todense()
                else:
                    x = temp_data.X
            else:
                if data_loc in temp_data.layers:
                    if issparse(temp_data.X):
                        x = temp_data.layers[data_loc].todense()
                    else:
                        x = temp_data.layers[data_loc]
                else:
                    raise ValueError(f"{data_loc} not 'X' or a valid layer in 'adata.layers'")
            scaler = StandardScaler()
            x_scaled = scaler.fit_transform(x).reshape(-1)
            density = KDEMultivariateConditional(
                endog=x_scaled,
                exog=adata.obsm[latent_rep],
                dep_type="c",
                indep_type="cc",
                bw="normal_reference",
                defaults=est_settings
            )
            z1 = density.cdf(thresh + np.zeros_like(x_scaled), adata.obsm[latent_rep])

            cm = plt.get_cmap(cmap)
            s = int(80000 / adata.shape[0]) if s is None else s

            # Plot density of gene expression
            idx = np.argsort(x_scaled)
            plt.subplot(nrows, ncols, index + 1)
            ax = sns.scatterplot(
                x=adata.obsm[latent_rep][idx, 0],
                y=adata.obsm[latent_rep][idx, 1],
                hue=1 - z1[idx],
                palette=cm,
                s=s,
                linewidth=0
            )
            ax.set_aspect("equal")  # Make sure the aspect ratio is square
            ax.set(xticklabels=[], yticklabels=[], xlabel=None, ylabel=None, title=gene)  # Get rid of x and y labels

            norm = plt.Normalize(1 - z1[idx].min(), 1 - z1[idx].max())
            sm = plt.cm.ScalarMappable(cmap=cm, norm=norm)
            sm.set_array([])

            # Remove the legend and add a colorbar
            ax.get_legend().remove()
            color_bar = ax.figure.colorbar(sm)
            color_bar.set_label("Probability of Expression")
    sns.despine(left=True, bottom=True)  # no borders
    fig.suptitle(title)  # figure title
    fig.tight_layout(rect=[0, 0.03, 1, 0.98])

    if save_path is not None:
        if "/" not in save_path:
            save_path = f"./{save_path}"
        check_path(save_path.rsplit("/", 1)[0])
        logger.info(f"Saving figure to {save_path}")
        plt.savefig(save_path, dpi=dpi, facecolor="white",)
    plt.show()



"""
Plots violinplot subplots of expression of genes
From scrnatools package

Created on Wed March 29 17:28:15 2023

@author: joe germino (joe.germino@ucsf.edu), nolan horner (nolan.horner@ucsf.edu)
"""
# external package imports
from anndata import AnnData
from typing import List, Optional, Tuple
import seaborn as sns
import matplotlib.pyplot as plt
import math
import scrnatools

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #

@debug(logger, configs)
def gene_violinplot(
    adata: AnnData,
    gene_list: List[str],
    x_key: str,
    layer: str = "X",
    x_values: List[str] = ['All'],
    hue_key: Optional[str] = None,
    hue_values: List[str] = ['All'],
    ncols: Optional[int] = None,
    nrows: Optional[int] = None,
    save_path: str = None,
    dpi: int = 300,
    *args, **kwargs
):
    """
    Plots violinplot subplots of expression of genes
    Parameters
    ----------
    adata
        The AnnData containing the gene expression and cell data.
    gene_list
        A list of genes to plot. Each gene will create a violinplot.
    x_key
        The categorical grouping to display on the x axis of the violinplot.
    layer
        The type of the expression data to use for the violinplot, can be a layer in 'adata.layers' or 'X'
        to use the data stored in adata.X. Default 'X'.
    x_values
        Values from x key group to display on violinplot. Default displays all values in x key.
    hue_key
        The categorical grouping to color the grouped violin plots by. Values will appear in legend. Default None.
    hue_values
        Values from hue_key to display on violinplot. Default displays all values in hue key.
    ncols
        Number of columns to display the violinplots.
    nrows
        Number of rows to display the violinplots.
    save_path
        The path to save the figure. Default None.
    dpi
        The resolution of the saved image. Default 300.

    Raises
    -------
    ValueError
        If a gene in 'gene_list' provided is not in provided AnnData layer
    ValueError
        If the 'x_values' provided less than 2.
    ValueError
        If the 'hue_values' is provided but hue_key is not provided.
    ValueError
        If the 'hue_values' provided less than 2 if hue_key is also provided.
    ValueError
        If ncols * nrows is less than length of gene list provided.
    ValueError
        If a value in 'x_values' provided is not in AnnData layer x key.
    ValueError
        If a value in 'hue_values' provided is not in AnnData layer hue key given hue key is also provided.
    """
    #check inputs
    invalid_genes = [i for i in gene_list if i not in adata.var_names]
    if len(invalid_genes) > 0:
        raise ValueError(f"Genes not found in data: {invalid_genes}")
    if len(x_values) < 2 and x_values != ['All']:
        raise ValueError(f"Please choose more than one x value. Or use '[All]' to see all {x_key}s")
    if hue_key is None:
        if hue_values != ['All']:
            raise ValueError(f"Please enter a hue_key.")
    else:
        if len(hue_values) < 2 and hue_values != ['All']:
            raise ValueError(f"Please choose more than one hue value. Or use '[All]' to see all {hue_key}s")
    if x_values == ['All']:
        x_values = adata.obs[x_key].unique()
    if hue_values == ['All'] and hue_key != None:
        hue_values = adata.obs[hue_key].unique()

    #format figure
    max_input_len = max(len(x_values), len(hue_values))
    min_input_len = min(len(hue_values), len(x_values))
    if (ncols is None) & (nrows is None):
        if max_input_len > 5:
            ncols = 1
            nrows = len(gene_list)
        elif max_input_len > 2:
            if min_input_len <= 2:
                ncols = 3
            else:
                ncols = 2
        else:
            if min_input_len <= 2:
                ncols = 5
            else:  
                ncols = 3
        nrows = int((len(gene_list) / ncols)) + 1 if (len(gene_list) % ncols) > 0 else int((len(gene_list) / ncols))
    elif (ncols is None) & (nrows is not None):
        ncols = int(len(gene_list) / nrows) + 1 if (len(gene_list) % nrows) > 0 else int((len(gene_list) / nrows))
    elif (nrows is None) & (ncols is not None):
        nrows = int((len(gene_list) / ncols)) + 1 if (len(gene_list) % ncols) > 0 else int((len(gene_list) / ncols))

    if ncols * nrows < len(gene_list):
            raise ValueError("Number of rows and columns must fit number of genes in gene list (nrows * ncols >= length of gene list).")
    
    fig=plt.figure(figsize=(ncols*len(x_values)*max(2, len(hue_values)) * 1, nrows*5))    
    
    #subset data on x values and hue values -> expression matrix
    if hue_key is None:
        subsetAdata = adata[adata.obs[x_key].isin(x_values), gene_list]
    else:
        subsetAdata = adata[adata.obs[x_key].isin(x_values) & adata.obs[hue_key].isin(hue_values), gene_list]
    expression_matrix = scrnatools.tl.get_expression_matrix(subsetAdata, gene_data = layer)
    
    expression_matrix[x_key] = subsetAdata.obs[x_key]
    if hue_key != None:
        expression_matrix[hue_key] = subsetAdata.obs[hue_key]
    
    #check inputted x values and hue values are in subsetted data
    for i in x_values:
        if i not in list(expression_matrix[x_key].unique()):
            raise ValueError(f"{i} is not in adata's x_key {x_key}.\nPossible {x_key}s: {list(adata.obs[x_key].unique())}")
    if hue_key != None:
        for i in hue_values:
            if i not in list(expression_matrix[hue_key].unique()):
                raise ValueError(f"{i} is not in adata's hue_key {hue_key}.\nPossible {hue_key}s: {list(adata.obs[hue_key].unique())}")

    #generate violin plots
    for plt_num, gene in enumerate(gene_list):
        plt.subplot(nrows, ncols, plt_num + 1)
        ax = sns.violinplot(
            expression_matrix,
            x=x_key,
            y=gene,
            hue=hue_key,
            scale='width',
            width = 0.9,
            order=x_values,
            hue_order=hue_values,
            *args, **kwargs
        )
        if hue_key is not None:
            if (plt_num > 0):
                plt.legend([],[], frameon=False)
            else:
                plt.legend(loc='upper right')
        plt.xticks(rotation=90)
        ax.grid(False)
        ax.set(xlabel=None)
        ax.set_title(f"{gene}")
    sns.set_theme(context="paper", style="white")
    fig.tight_layout()
    if save_path is not None:
        if "/" not in save_path:
            save_path = f"./{save_path}"
        check_path(save_path.rsplit("/", 1)[0])
        logger.info(f"Saving figure to {save_path}")
        plt.savefig(save_path, dpi=dpi, facecolor="white",)
    plt.show()



"""
Loads the lookup tables for an isoform AnnData object
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
import json
from typing import Tuple, Dict

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def load_isoform_lookup_tables(path: str) -> Tuple[Dict[str, str], Dict[str, str], Dict[str, str]]:
    """
    Loads the lookup tables for an isoform AnnData object

    Parameters
    ----------
    path
        The path and file name of the lookup dicts to load. (_ec2tx.json, _ec2g.json, and _inv_map.json all appended to
        this file name)

    Returns
    -------
    ec2tx
        The equivalence class to transcript dict
    ec2g
        The equivalence class to gene dict
    inv_map
        The gene to equivalence class dict
    """
    with open(f"{path}_ec2tx.json") as d:
        ec2tx = json.load(d)
    with open(f"{path}_ec2g.json") as d:
        ec2g = json.load(d)
    with open(f"{path}_inv_map.json") as d:
        inv_map = json.load(d)
    return ec2tx, ec2g, inv_map



"""
Extracts the cell x gene expression matrix from an AnnData object
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
from typing import Optional
from scipy.sparse import issparse
from pandas import DataFrame
from anndata import AnnData

# scrnatools package imports
from .._configs import configs
from .._utils import debug

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def get_expression_matrix(
        anndata: AnnData,
        gene_data: Optional[str] = None,
) -> DataFrame:
    """
    Extracts the cell x gene expression matrix from an AnnData object

    Parameters
    ----------
    anndata
        The AnnData to pull the expression matrix from
    gene_data
        Where to get the expression data from the AnnData object. If a layer from the AnnData['layers'] is passed,
        that is used, otherwise 'X' or 'raw' can be used.

    Returns
    -------
    A Pandas DataFrame containing the expression matrix (cells x genes)

    Raises
    -------
    ValueError
        When 'layer' is not one of the layers in 'anndata'
    """
    if gene_data == "X":
        # Convert to a dense matrix if needed
        if issparse(anndata.X):
            matrix = DataFrame(anndata.X.todense(), index=anndata.obs.index, columns=anndata.var_names)
        else:
            matrix = DataFrame(anndata.X, index=anndata.obs.index, columns=anndata.var_names)
    # get a layer from the AnnData if specified
    elif gene_data == "raw":
        raw_data = anndata.raw.to_adata()
        # Convert to a dense matrix if needed
        if issparse(raw_data.X):
            matrix = DataFrame(raw_data.X.todense(), index=anndata.obs.index, columns=anndata.raw.var_names)
        else:
            matrix = DataFrame(raw_data.X, index=anndata.obs.index, columns=anndata.raw.var_names)
    elif gene_data in anndata.layers:
        if issparse(anndata.layers[gene_data]):
            matrix = DataFrame(anndata.layers[gene_data].todense(), index=anndata.obs.index, columns=anndata.var_names)
        else:
            matrix = DataFrame(anndata.layers[gene_data], index=anndata.obs.index, columns=anndata.var_names)
    else:
        raise ValueError(f"{gene_data} is not 'X', 'raw', or a valid layer name in '{anndata.layers}'")
    return matrix



"""
Creates a UMAP/TSNE plot of a gene's expression from a layer using scanpy
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import scipy
from typing import List, Tuple
from anndata import AnnData

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def gene_embedding(
        adata: AnnData,
        gene_list: List[str],
        layer: str,
        figsize: Tuple[int, int] = (4, 4),
        dpi: int = 80,
        dpi_save: int = 300,
        min_quantile: float = 0.01,
        max_quantile: float = 0.99,
        use_rep="X_umap",
        *args,
        **kwargs,
):
    """
    Creates a UMAP plot of a gene's expression from a layer using scanpy
    Parameters
    ----------
    adata
        The adata containing expression data
    gene_list
        The list of genes to plot
    layer
        The layer containing expression data, or 'X' to use the expression data stored in adata.X
    figsize
        The size of each figure. Default (4, 4)
    dpi
        The dpi of plots shown inline. Default 80
    dpi_save
        The dpi of saved plots. Default 300
    min_quantile
        The quantile of expression for a gene to set the minimum of the colorbar to. Default 0.01
    max_quantile
        The quantile of expression for a gene to set the maximum of the colorbar to. Default 0.99
    use_rep
        The embedding coordinates in 'adata.obsm' to use, default is standard UMAP representation from scanpy
    args
        Other arguments to 'sc.pl.umap()'
    kwargs
        Other arguments to 'sc.pl.umap()'

    Raises
    -------
    ValueError
        If 'layer' is not a valid key in 'adata.layers'
    """
    # Setup figure
    sc.set_figure_params(figsize=figsize, dpi=dpi, dpi_save=dpi_save, facecolor="white", frameon=False)
    sns.set_context("paper")
    plt.rcParams["axes.grid"] = False
    vmin = []
    vmax = []
    if layer != "X" and layer not in adata.layers.keys():
        raise ValueError(f"{layer} is not 'X' or a valid key in 'adata.layers'")
    # Get min/max expression values for the colorbars for each gene
    for gene in gene_list:
        if layer != "X":
            if scipy.sparse.issparse(adata.layers[layer]):
                min = np.quantile(adata[:, gene].layers[layer].todense().tolist(), min_quantile)
                max = np.quantile(adata[:, gene].layers[layer].todense().tolist(), max_quantile)
                vmin.append(min)
                if max < 1:
                    vmax.append(1)
                else:
                    vmax.append(max)
            else:
                min = np.quantile(adata[:, gene].layers[layer], min_quantile)
                max = np.quantile(adata[:, gene].layers[layer], max_quantile)
                vmin.append(min)
                vmax.append(max)
        else:
            min = np.quantile(adata[:, gene].X, min_quantile)
            max = np.quantile(adata[:, gene].X, max_quantile)
            vmin.append(min)
            vmax.append(max)
    # Plot the data
    adata_copy = adata.copy()
    adata_copy.obsm["X_umap"] = adata.obsm[use_rep]
    if layer == "X":
        sc.pl.umap(
            adata_copy,
            color=gene_list,
            vmin=vmin,
            vmax=vmax,
            *args,
            **kwargs,
        )
    else:
        sc.pl.umap(
            adata_copy,
            color=gene_list,
            layer=layer,
            vmin=vmin,
            vmax=vmax,
            *args,
            **kwargs,
        )



"""
Tools init script
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
from ._get_expression_matrix import get_expression_matrix
from ._cluster_de import cluster_de
from ._log_density_ratio import log_density_ratio
from ._cell_type_similarity import cell_type_similarity
from ._create_cell_type_signature import create_cell_type_signature
from ._get_immgen_similarity_signatures import get_immgen_similarity_signatures
from ._read_kallisto_tcc_matrix import read_kallisto_tcc_matrix
from ._isoform_preprocessing import isoform_preprocessing
from ._create_isoform_lookup_tables import create_isoform_lookup_tables
from ._save_isoform_lookup_tables import save_isoform_lookup_tables
from ._load_isoform_lookup_tables import load_isoform_lookup_tables



"""
Calculates the cosine similarity of cells to provided cell type signatures
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
from anndata import AnnData
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import StandardScaler
from typing import Optional
import pandas as pd

# scrnatools package imports
from .._configs import configs
from .._utils import debug

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def cell_type_similarity(
        adata: AnnData,
        signatures: pd.DataFrame,
        normalize: str = "cell",
) -> AnnData:
    """
    Calculates the cosine similarity of cells to provided cell type signatures

    Parameters
    ----------
    adata
        The AnnData object containing cells to score, with library size rescaled and log-normalized counts stored in
        'adata.raw'
    signatures
        A DataFrame containing transcriptome signatures for each cell type to be tested. Columns are cell types and rows
        are the log-normalized expression of each gene in that cell type. Can be created from cell type clusters of a
        scRNAseq dataset using 'create_cell_type_signature'
    normalize
        How to rescale the cosine similarity scores. Default is 'cell', which calculates a z-score normalized for each
        cell across all cell types in 'signatures'. Other possible values are 'cell type' which calculates a z-score
        normalized for each cell type in 'signatures' across all cells and 'none' which returns the raw cosine
        similarity values.

    Returns
    -------
    The AnnData passed in with 'adata' with the cosine similarity scores for each immgen cell type added as a column to
    'adata.obs'
    """

    # Raw attribute contains the log-normalized counts
    cell_data = pd.DataFrame(adata.raw.X.toarray(), columns=adata.raw.var_names, index=adata.obs.index).T
    cell_data = cell_data / cell_data.sum(axis=0)  # Normalize per cell expression to sum to 1
    # Join with gene x cell type signature matrix so that only shared genes are kept
    all_df = cell_data.join(
        signatures,
        how="inner"
    )
    all_df = all_df / all_df.sum(axis=0)  # Normalize per cell expression to sum to 1

    # Calculate cosine similarity of single cells to each cell type
    sim = cosine_similarity(
        all_df[all_df.columns[:cell_data.shape[1]]].values.T,  # Single-cell expression data
        all_df[all_df.columns[cell_data.shape[1]:]].values.T  # Cell type gene signature expression data
    )

    # Create dataframe with per cell similarity scores for each cell type
    similarity = pd.DataFrame(
        sim,
        columns=all_df.columns[cell_data.shape[1]:],
        index=all_df.columns[:cell_data.shape[1]]
    )
    if normalize == "cell type":
        logger.info(f"Scaling cosine similarity scores to z-scores by cell type")
        # Scale similarity scores
        scaler = StandardScaler()
        minmax_scale = scaler.fit(similarity)
        x_minmax = minmax_scale.transform(similarity)
        scale_by_row = pd.DataFrame(x_minmax)
    elif normalize == "cell":
        logger.info(f"Scaling cosine similarity scores to z-scores by cell")
        # Scale similarity scores
        scaler = StandardScaler()
        minmax_scale = scaler.fit(similarity.T)
        x_minmax = minmax_scale.transform(similarity.T).T
        scale_by_row = pd.DataFrame(x_minmax)
    else:
        logger.info(f"Leaving similarity scores unscaled")
        scale_by_row = pd.DataFrame(similarity)

    # Add cosine similarity scores to adata obs
    scale_by_row.index = similarity.index
    if normalize != "none":
        scale_by_row.columns = similarity.columns + "_cosine_similarity_" + normalize + "_z_score"
    else:
        scale_by_row.columns = similarity.columns + "_cosine_similarity"
    adata_copy = adata.copy()
    adata_copy.obs = adata_copy.obs.join(scale_by_row)
    return adata_copy



"""
Calculates DE marker genes for data clusters
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
from anndata import AnnData
from pandas import DataFrame
from typing import Optional, Dict
from scvi.model import SCVI

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def cluster_de(
        adata: AnnData,
        model: SCVI,
        cluster_key: str,
        lfc_mean_threshold: int = 0,
        bayes_factor_threshold: int = 3,
        non_zeroes_proportion_threshold: int = 0.1,
        save_path: Optional[str] = None,
        subset: bool = True,
) -> Dict[str, DataFrame]:
    """
    Calculates DE marker genes for data clusters

    Parameters
    ----------
    adata
        The data to analyze
    model
        The scVI model for 'adata'
    cluster_key
        The column name of the cluster data in 'adata.obs'
    lfc_mean_threshold
        The minimum lfc_mean to filter de genes on (exclusive). Default 0
    bayes_factor_threshold
        The minimum bayes factor to filter de genes on (exclusive). Default 3
    non_zeroes_proportion_threshold
        The minimum proportion of cells with non-zero expression filter de genes on (exclusive). Default 0.1
    save_path
        The path to save the marker gene lists to. Default None
    subset
        Whether to subset the DE gene list based on thresholds or return all genes in data. Default True

    Returns
    -------
    A dictionary with keys equal to the categories of the cluster column in 'adata.obs' (i.e. cell types) linked to
    DataFrames of the filtered (or unfiltered but sorted) DE marker genes for that cluster

    Raises
    -------
    ValueError
        If 'cluster_key' is not a valid column in 'adata.obs.columns'
    """
    if cluster_key not in adata.obs.columns:
        raise ValueError(f"{cluster_key} is not a valis column in 'adata.obs'")
    de_df = model.differential_expression(groupby=cluster_key, )
    cats = adata.obs[cluster_key].cat.categories
    markers = {}
    for i, c in enumerate(cats):
        cid = f"{c} vs Rest"
        cell_type_df = de_df.loc[de_df.comparison == cid]
        if subset:
            cell_type_df = cell_type_df[cell_type_df.lfc_mean > lfc_mean_threshold]
            cell_type_df = cell_type_df[cell_type_df["bayes_factor"] > bayes_factor_threshold]
            cell_type_df = cell_type_df[cell_type_df["non_zeros_proportion1"] > non_zeroes_proportion_threshold]
        markers[c] = cell_type_df
    if save_path is not None:
        check_path(save_path)
        for cluster, marker_genes in markers.items():
            if save_path[-1] == "/":
                save_path = save_path[:-1]
            check_path(save_path)
            marker_genes.to_csv(f"{save_path}/{cluster}.csv")
    return markers



"""
sc-rna-tools package global settings manager
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
import logging
import os

# <name TBD> package imports
from ._utils import type_check


class Config:
    """Config manager for scrnatools"""

    def __init__(
            self,
            verbosity="warning",
            debug_memory=False,
            debug_timing=False,
            log_path="sc-rna-tools_logs",
            save_logs=False,
    ):
        self._verbosity = verbosity
        self._debug_memory = debug_memory
        self._debug_timing = debug_timing
        self._log_path = log_path
        self._save_logs = save_logs
        self._loggers = []  # list of loggers created by modules

    @property
    def verbosity(self) -> str:
        return self._verbosity

    @verbosity.setter
    def verbosity(self, verbosity: str):
        """Sets the verbosity level of the logger

        Allowed are: 'debug', 'info', 'warning', 'error', 'critical'
        """
        type_check(verbosity, "verbosity", str)
        # set the level of all the loggers
        for logger in self._loggers:
            self._set_log_level(logger, verbosity)
        self._verbosity = verbosity

    @property
    def debug_memory(self) -> bool:
        """Determines whether line by line memory usage is logged for debugging functions"""
        return self._debug_memory

    @debug_memory.setter
    def debug_memory(self, debug_memory: bool):
        type_check(debug_memory, "debug_memory", bool)
        self._debug_memory = debug_memory

    @property
    def debug_timing(self) -> bool:
        """Determines whether function timing is logged for debugging"""
        return self._debug_timing

    @debug_timing.setter
    def debug_timing(self, debug_timing: bool):
        type_check(debug_timing, "debug_timing", bool)
        self._debug_timing = debug_timing

    @property
    def log_path(self) -> str:
        """The path where log files are saved"""
        return self._log_path

    @log_path.setter
    def log_path(self, path: str):
        type_check(path, "path", str)
        self._log_path = path

    @property
    def save_logs(self) -> bool:
        """Determines whether logging messages are saved to a log file"""
        return self._save_logs

    @save_logs.setter
    def save_logs(self, save_logs: bool):
        type_check(save_logs, "save_logs", bool)
        # if logs are not currently being written to a file, but being updated to be written to a file, add a
        # FileHandler to each logger
        if not self.save_logs and save_logs:
            # make sure log directory exists
            self.check_log_path()
            # create the logging format and FileHandler
            formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s: %(message)s", "%Y-%m-%d %H:%M:%S")
            file_handler = logging.FileHandler(f"{self.log_path}/sc-rna-tools.log")
            file_handler.setFormatter(formatter)
            # add the FileHandler to each logger if it doesn't already have a FileHandler
            for logger in self._loggers:
                handlers = logger.handlers
                if sum([isinstance(h, logging.FileHandler) for h in handlers]) == 0:  # check if there is a FileHandler
                    logger.addHandler(file_handler)
        # if logs are currently being written to a file, but being updated to not be written to a file, remove the
        # FileHandler to each logger
        if self.save_logs and not save_logs:
            for logger in self._loggers:
                # find the FileHandler for this logger and remove it
                for h in logger.handlers:
                    if isinstance(h, logging.FileHandler):
                        logger.removeHandler(h)
        self._save_logs = save_logs

    def __str__(self) -> str:
        return f"ScoreConfig(" \
               f"verbosity: {self.verbosity}, " \
               f"debug_memory: {self.debug_memory}, " \
               f"debug_timing: {self.debug_timing}, " \
               f"log_path: {self.log_path}, " \
               f"save_logs: {self.save_logs})"

    def clear_logs(self):
        """Clears all logs saved in the logs path"""
        if os.path.exists(self._log_path):
            for f in os.listdir(self._log_path):
                # Remove all the files in the log directory
                os.remove(os.path.join(self._log_path, f))
            # remove the log directory
            os.rmdir(self._log_path)

    def check_log_path(self):
        """Checks if the log path exists, and creates it if it doesn't"""
        if not os.path.exists(self._log_path):
            os.makedirs(self._log_path)

    def create_logger(self, name: str):
        """Called by modules to create a logger for that function

        Parameters
        ----------
        name
            The name of the logger (typically the module name creating the logger)
        """
        type_check(name, 'name', str)
        # create a logger, StreamHandler, and the formatter for that logger
        logger = logging.getLogger(name)
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s: %(message)s", "%Y-%m-%d %H:%M:%S")
        handler.setFormatter(formatter)
        handlers = logger.handlers
        if sum([isinstance(h, logging.StreamHandler) for h in handlers]) == 0:  # check if there is a StreamHandler
            logger.addHandler(handler)
        # if logs are being saved to a file as well, create a FileHandler for the new logger too
        if self.save_logs:
            self.check_log_path()
            file_handler = logging.FileHandler(f"{self.log_path}/sc-rna-tools.log")
            file_handler.setFormatter(formatter)
            if sum([isinstance(h, logging.FileHandler) for h in handlers]) == 0:  # check if there is a StreamHandler
                logger.addHandler(file_handler)
        # Set the log level to the current verbosity
        self._set_log_level(logger, self.verbosity)
        # Add the logger to the master list
        self._loggers.append(logger)
        return logger

    @staticmethod
    def _set_log_level(logger: logging.Logger, level: str):
        """Sets the log level of a modules logger

        Parameters
        ----------
        logger
            The logger to set the level of
        level
            String corresponding to the name of the logging level to set.
            Allowed: 'info', 'debug', 'warning', 'error', 'critical'
        """
        log_levels = {
            "info": logging.INFO,
            "debug": logging.DEBUG,
            "warning": logging.WARNING,
            "error": logging.ERROR,
            "critical": logging.CRITICAL
        }
        if level not in log_levels:
            raise ValueError(f"'{level}' is not a valid logging level "
                             f"('info', 'debug', 'warning', 'error', 'critical')")
        logger.setLevel(log_levels[level])


configs = Config()



"""
Fills the default value for parameter that don't set values for all plots (keys)
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
from typing import Any, List

# -------------------------------------------------------function----------------------------------------------------- #


def fill_array(variable: Any, default: Any, final_size: int) -> List[Any]:
    """
    Fills the default value for parameter that don't set values for all plots (keys)

    Parameters
    ----------
    variable
        The parameter to fill values for
    default
        The default value for the argument on a single plot
    final_size
        The number of keys in the figure

    Returns
    -------
    A list of length 'final_size' with the default values for the parameter filling to the end of the list
    """
    # ensure the parameter is a list
    if not isinstance(variable, list):
        variable = [variable]
    # get the current length of the parameter (i.e. the number of keys with user-specified values)
    num_vars = len(variable)
    if num_vars < final_size:
        for _ in range(final_size - num_vars):
            if num_vars == 1:
                # if only one value is provided for the parameter, use that by default for all keys
                variable.append(variable[0])
            else:
                # otherwise, add the default value for the parameter to fill the list
                variable.append(default)
    return variable



"""
Utils init script
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# import functions from their modules

from ._type_check import type_check
from ._debug import debug
from ._fill_array import fill_array
from ._check_path import check_path



"""
Performs basic preprocessing of kallisto isoform data
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
from anndata import AnnData
import scanpy as sc
from pandas import DataFrame
import numpy as np

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def isoform_preprocessing(
        adata: AnnData,
        ec: DataFrame,
):
    """
    Performs basic preprocessing of kallisto isoform data

    Parameters
    ----------
    adata
        The AnnData containing kallisto isoform data
    ec
        The equivalence class matrix to rename vars with
    """
    adata.obs_names_make_unique()
    # Filter out equivalence classes with no counts
    ngenes = adata.shape[1]
    sc.pp.filter_genes(adata, min_counts=1)
    logger.info(f"Filtered {ngenes - adata.shape[1]} ecs with no counts")
    # Filter cells with no ecs
    ncells = adata.shape[0]
    sc.pp.filter_cells(adata, min_genes=1)
    logger.info(f"Filtered {ncells - adata.shape[0]} cells with no ecs")
    # rename vars to be ecs
    ec = ec.iloc[np.array(adata.var_names, dtype=int)]
    adata.var_names = ec.loc[:, 1].values



"""
Creates a csv with the immgen cell populations signature data
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
import pandas as pd
from typing import Optional
import subprocess
import os
from shutil import which

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def get_immgen_similarity_signatures(
        save_path: str = "datasets"
) -> pd.DataFrame:
    """
    Creates a csv with the immgen cell populations signature data

    Parameters
    ----------
    save_path
        The path to save the immgen data and signature file to

    Returns
    -------
        A DataFrame containing the immgen cell type expression signatures, with cell types as columns and genes as rows

    Raises
    -------
    OSError
        if wget is not installed on the system
    """

    if save_path[-1] == "/":
        save_path = save_path[:-1]
    # Download data from github to datasets folder
    if which("wget") is None:
        raise OSError("wget is required to download files. Please install using 'sudo apt-get install wget' on UNIX or "
                      "download and install from 'http://gnuwin32.sourceforge.net/packages/wget.htm' on Windows")
    if not os.path.exists(f"{save_path}/immgen"):
        check_path(save_path)
        subprocess.run(["wget", "-P", f"{save_path}/",
                        "https://gist.github.com/vasilisNt/5e23eeefc188e1e772f428c74ef43277/raw/67f83d282b0b2180a8eeff74edf079d8826b12ba/immgen.tar.gz"],
                       stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL,)
        subprocess.run(["tar", "-xzf", f"{save_path}/immgen.tar.gz", "-C", f"{save_path}/"])
        subprocess.run(["rm", "-rf", f"{save_path}/immgen.tar.gz"])

    # Import immgen data
    immgen = pd.read_csv(f"{save_path}/immgen/Immgen_expression.txt", sep='\t', index_col=0)
    probes = pd.read_csv(f"{save_path}/immgen/Immgen_probes.txt", sep='\t', index_col=0, header=None)
    g2n = pd.read_csv(f"{save_path}/immgen/mart_export.txt", sep='\t', )

    # create DataFrame with signatures
    probes = probes.join(g2n.set_index('Gene stable ID')['Gene name'], on=1)
    probes.columns = ['gene_id', 'gene_name']
    immgen = immgen.join(probes['gene_name']).set_index('gene_name')
    immgen = immgen.groupby(immgen.index).aggregate('sum')

    # Save signatures csv
    immgen.to_csv(f"{save_path}/immgen/immgen_signatures.csv")

    return immgen



"""
Checks the type of a variable
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
from typing import Any

# -------------------------------------------------------function----------------------------------------------------- #


def type_check(var: Any, varname: str, types: Any,):
    """
    Checks the type of a variable

    Parameters
    ----------
    var
        the variable to check the type of
    varname
        the name of the variable
    types
        the type the variable should be

    Raises
    -------
    TypeError
        when 'var' is not one of 'types'
    """
    if not isinstance(var, types):
        raise TypeError(f"{varname} must be of type {types}")



"""
Plots a heatmap of expression of genes
From scrnatools package

Created on Tue March 28 17:47:46 2023

@author: joe germino (joe.germino@ucsf.edu), nolan horner (nolan.horner@ucsf.edu)
"""
# external package imports
from anndata import AnnData
from typing import List, Optional, Tuple
import seaborn as sns
import matplotlib.pyplot as plt
import scrnatools

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #

@debug(logger, configs)
def gene_heatmap(
    adata: AnnData,
    gene_list: List[str],
    obs_key: str,
    layer: str = "X",
    obs_values: List[str] = ['All'],
    cbar_args: List[int] = None,
    reverse_axis: bool = False,
    cell_size: int = 50,
    save_path: str = None,
    dpi: int = 300,
    *args, **kwargs
):
    """
    Plots a heatmap of expression of genes
    Parameters
    ----------
    adata
        The AnnData containing the gene expression and cell data.
    gene_list
        A list of genes to plot.
    obs_key
        The categorical grouping to display on the heatmap. Can use 'cell_type' to display gene
        expression for different cell types.
    layer
        The type of the expression data to use for the heatmap, can be a layer in 'adata.layers' or 'X'
        to use the data stored in adata.X. Default 'X'.
    obs_values
        Values from obs key group to display on heatmap. Default displays all values in obs key.
    cbar_args
        List of integers to position color bar on heatmap. [x position, y position, width, height].
        Default automatically adjusts to right middle of heatmap.
    reverse_axis
        A boolean to reverse x and y axes. Default False.
    cell_size
        Integer specifying the size of each cell in the heatmap. Default 50.
    save_path
        The path to save the figure. Default None.
    dpi
        The resolution of the saved image. Default 300.

    Raises
    -------
    ValueError
        If a gene in 'gene_list' provided is not in provided AnnData layer.
    ValueError
        If 'gene_list' provided less than 2.
    ValueError
        If the 'obs_values' provided less than 2.
    ValueError
        If a value in 'obs_values' provided is not in provided AnnData layer obs key
    """
    invalid_genes = [i for i in gene_list if i not in adata.var_names]
    if len(invalid_genes) > 0:
        raise ValueError(f"Genes not found in data: {invalid_genes}")
    if len(gene_list) < 2:
        raise ValueError(f"Please choose two or more genes. {len(gene_list)} genes were entered.")
    if layer == "raw":
        raw_adata = adata.raw.to_adata()
        expression_matrix = scrnatools.tl.get_expression_matrix(raw_adata[:, gene_list], gene_data = "X")
    else:
        expression_matrix = scrnatools.tl.get_expression_matrix(adata[:, gene_list], gene_data = layer)
    expression_matrix[obs_key] = adata.obs[obs_key]
    expression_matrix = expression_matrix.groupby(obs_key).mean() #matrix with mean expression levels of selected genes for each cell type
    if obs_values != ['All']: #include only cell types from cell_type_list
        if len(obs_values) < 2:
            raise ValueError(f"Please choose more than one obs value. Or use '[All]' to see all {obs_key}s")
        for i in obs_values:
            if i not in expression_matrix.index:
                raise ValueError(f"{i} is not in adata's obs_key {obs_key}.\nPossible {obs_key}s: {list(expression_matrix.index.categories)}")
        for j in expression_matrix.index:
            if j not in obs_values:
                expression_matrix = expression_matrix.drop(j)

    if reverse_axis:
        expression_matrix = expression_matrix.transpose()
        
    figdpi = plt.rcParams['figure.dpi']
    totalWidth = plt.rcParams['figure.subplot.right']-plt.rcParams['figure.subplot.left']
    totalHeight = plt.rcParams['figure.subplot.top']-plt.rcParams['figure.subplot.bottom']
    nrows, ncols = expression_matrix.shape
    figWidth = (ncols*cell_size/figdpi)/totalWidth
    figHeight = (nrows*cell_size/figdpi)/totalHeight
    
    cg = sns.clustermap(
        expression_matrix, figsize = (figWidth, figHeight),
        *args, **kwargs
    )
    
    # calculate the size of the heatmap axes
    axWidth = (ncols*cell_size)/(figWidth*figdpi)
    axHeight = (nrows*cell_size)/(figHeight*figdpi)

    # resize heatmap
    ax_heatmap_og_pos = cg.ax_heatmap.get_position()
    cg.ax_heatmap.set_position([ax_heatmap_og_pos.x0, ax_heatmap_og_pos.y0, 
                                  axWidth, axHeight])
    
    #layout of colorbar
    if cbar_args is None:
        cbar_x_pos = 0.85 + ax_heatmap_og_pos.x0
        cbar_y_pos = ax_heatmap_og_pos.y0
        cbar_width = 25/(figWidth*figdpi)
        cbar_height = axHeight
        cbar_args = [cbar_x_pos, cbar_y_pos, cbar_width, cbar_height]

    #set the layout
    cg.ax_cbar.set_position(cbar_args)
    cg.ax_cbar.grid(False)
    cg.ax_row_dendrogram.set_visible(False)
    cg.ax_col_dendrogram.set_visible(False)
    cg.ax_heatmap.yaxis.tick_left()
    cg.ax_heatmap.set_ylabel("")
    cg.ax_heatmap.set_xlabel("")
    cg.ax_heatmap.grid(False)
    cg.ax_heatmap.tick_params(axis='y', labelrotation=0)
    cg.ax_heatmap.tick_params(axis='x', labelrotation=90)

    sns.set_theme(context="paper", style="white")

    if save_path is not None:
        if "/" not in save_path:
            save_path = f"./{save_path}"
        check_path(save_path.rsplit("/", 1)[0])
        logger.info(f"Saving figure to {save_path}")
        plt.savefig(save_path, dpi=dpi, facecolor="white", bbox_inches="tight")
    plt.show()



"""
Saves the lookup tables for an isoform AnnData object
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
import json
from typing import Dict

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def save_isoform_lookup_tables(
        path: str,
        ec2tx: Dict[str, str],
        ec2g: Dict[str, str],
        inv_map: Dict[str, str],
):
    """
    Saves the lookup tables for an isoform AnnData object

    Parameters
    ----------
    path
        The path and file name to save the lookup dicts to. (_ec2tx.json, _ec2g.json, and _inv_map.json all appended to
        this file name)
    ec2tx
        The equivalence class to transcript dict to save
    ec2g
        The equivalence class to gene dict to save
    inv_map
        The gene to equivalence class dict to save
    """
    with open(f"{path}_ec2tx.json", "w") as f:
        json_file = json.dumps(ec2tx)
        f.write(json_file)
    with open(f"{path}_ec2g.json", "w") as f:
        json_file = json.dumps(ec2g)
        f.write(json_file)
    with open(f"{path}_inv_map.json", "w") as f:
        json_file = json.dumps(inv_map)
        f.write(json_file)



"""
QC init script
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

from ._filter_cells import filter_cells
from ._scrublet import scrublet



"""
Reads in a kallisto bustools isoform matrix
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
import scanpy as sc
import pandas as pd
from pandas import DataFrame
from typing import Tuple
from anndata import AnnData
import numpy as np

# scrnatools package imports
from .._configs import configs
from .._utils import debug

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def read_kallisto_tcc_matrix(
        sample_path: str,
) -> Tuple[AnnData, DataFrame]:
    """
    Reads in a kallisto bustools isoform matrix

    Parameters
    ----------
    sample_path
        The path to and sample name for the mtx, barcodes, and ec files from kallisto bustools output

    Returns
    -------
        An AnnData with the tcc data and the ec matrix
    """
    logger.info(f"Reading sample: {sample_path}")
    data = sc.read_mtx(f"{sample_path}.mtx")
    data.obs_names = pd.read_csv(f"{sample_path}.barcodes.txt", header=None).values.T[0]
    ec = pd.read_csv(f"{sample_path}.ec.txt", sep='\t', header=None)
    data.var_names = np.array(np.arange(data.shape[1]), dtype=str)
    return data, ec



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scrnatools",
    version="0.3.5",
    author="Joe Germino",
    author_email="joe.germino@ucsf.edu",
    description="Tools for single cell RNA sequencing pipelines",
    url="https://github.com/j-germino/sc-rna-tools",
    packages=setuptools.find_packages(),
    install_requires=[
        "scanpy",
        "scrublet",
        "scvi-tools",
        "matplotlib",
        "pandas",
        "numpy",
        "seaborn",
        "memory-profiler",
        "leidenalg",
        "datetime",
        "anndata",
        "scipy",
    ]
)



"""
Checks if a path exists and creates it if it doesn't
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
import os

# -------------------------------------------------------function----------------------------------------------------- #


def check_path(path: str):
    """
    Checks if a path exists and creates it if it doesn't

    Parameters
    ----------
    path
        The string of the path to check/create
    """
    if not os.path.exists(path):
        os.makedirs(path)



"""
Decorator function to handle debugging
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
import datetime
from logging import Logger
from memory_profiler import profile
from time import perf_counter
from typing import Any

# -------------------------------------------------------function----------------------------------------------------- #


def debug(logger: Logger, configs,) -> Any:
    """A function decorator for creating various debugging logs"""

    def decorator(func):
        def wrapper(*args, **kwargs):
            # log the function call
            logger.debug(f"Function call of {func.__name__}")  # with args: {args}, and kwargs: {kwargs}")
            start = perf_counter()
            # if tracking memory usage line by line
            if configs.debug_memory:
                # check/make the log directory
                configs.check_log_path()
                # get datetime string for unique memory usage log file
                current_time = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S.%f")
                # wrap the function with the memory profiler, saving to a log file
                with open(f"{configs.log_path}/{current_time}_{func.__name__}_memory.txt", "w+") as file:
                    f = profile(func, stream=file)
                    # execution of the function to wrap
                    result = f(*args, **kwargs)
            else:
                # execution of the function to wrap
                result = func(*args, **kwargs)
            end = perf_counter()
            # if tracking time to execute
            if configs.debug_timing:
                minutes = int((end - start) / 60)
                seconds = (end - start - minutes * 60)
                # log time to execute as XX min X.XXXX s
                logger.info(f"{func.__name__} took: {minutes:d} min {seconds:.4f} s")
            return result

        return wrapper

    return decorator



"""
Plots qc metrics for kallisto isoform data
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
import matplotlib.pyplot as plt
import seaborn as sns
import scanpy as sc
import numpy as np
from anndata import AnnData
from typing import Optional
from matplotlib.patches import Patch

# scrnatools package imports
from scrnatools._configs import configs
from scrnatools._utils import debug

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def isoform_qc_plot(
        adata: AnnData,
        batch_key: Optional[str] = None,
):
    """
    Plots qc metrics for kallisto isoform data

    Parameters
    ----------
    adata
        The AnnData containing isoform data to plot
    batch_key
        An obs column identifying batches to plot separately

    Raises
    ----------
    ValueError
        If the batch key provided is not a column in in 'adata.obs'
    """
    if batch_key not in adata.obs.columns:
        raise ValueError(f"{batch_key} is not a valid column in 'adata.obs'")

    sc.set_figure_params(figsize=(6, 6), dpi=80, dpi_save=300, facecolor="white", frameon=False)
    sns.set_context("notebook")
    plt.rcParams["axes.grid"] = False

    colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22',
              '#17becf']
    if len(adata.obs[batch_key].unique()) > len(colors):
        logger.info(f"Too many batches in adata.obs.{batch_key}, colors will overlap")
    if batch_key is None:
        adata.obs["qc_plot_batch"] = "all_cells"
        batch_key = "qc_plot_batch"

    # Counts per ec plot
    fig = plt.figure(figsize=(12, 6))
    fig.add_subplot(1, 2, 1)
    plt.semilogy(np.sort(adata.var['total_counts'].values)[::-1])
    plt.ylabel("total_counts")
    plt.xlabel("ec")

    # Knee plot (all ec counts per barcode)
    fig.add_subplot(1, 2, 2)
    legend_elements = []
    for i, batch in enumerate(adata.obs[batch_key].unique()):
        q = np.sort(np.array(adata[adata.obs[batch_key] == batch].X.sum(axis=1)).reshape(-1))[::-1]
        plt.loglog(q, linewidth=1)
        plt.ylabel("UMI Counts")
        plt.xlabel("Barcodes")
        legend_elements.append(Patch(facecolor=colors[i], label=batch))

    plt.legend(handles=legend_elements)
    plt.show()



"""
Creates the lookup tables for isoform data transcripts and genes
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
from anndata import AnnData
from pandas import DataFrame
import numpy as np
import pandas as pd
from typing import Tuple, Dict

# scrnatools package imports
from .._configs import configs
from .._utils import debug, check_path

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def create_isoform_lookup_tables(
        adata: AnnData,
        t2enst_path: DataFrame,
        t2g_path: DataFrame,
) -> Tuple[Dict[str, str], Dict[str, str], Dict[str, str]]:
    """
    Creates the lookup tables for isoform data transcripts and genes

    Parameters
    ----------
    adata
        The AnnData containing kallisto isoform data
    t2enst_path
        Path to the transcript to ensembl transcript id mapping dataframe (from kallisto alignment)
    t2g_path
        Path to the transcript to gene mapping dataframe (from the kallisto reference used for alignment)

    Returns
    ----------
    ec2tx
        The equivalence class to transcript dict
    ec2g
        The equivalence class to gene dict
    inv_map
        The gene to equivalence class dict
    """
    # Import the transcript list from kallisto
    t2enst = pd.read_csv(t2enst_path, header=None,)
    t2enst.columns = ['enst']

    # Import the transcript to gene mapping dataframe
    t2g = pd.read_csv(t2g_path, header=None, sep='\t')
    t2g.columns = ['enst', 'ensg', 'gname']
    t2g.index = t2g.enst

    # Get ecs from adata
    ecs = np.array(adata.var_names.values, dtype=str)
    # Convert ec string to a list of tx indices
    ecdict = {}
    for ec, txs in enumerate(ecs):
        ecdict[str(ec)] = list(np.array(txs.split(','), dtype=int))
    # Save ec list to adata
    adata.var['ecs'] = ecs
    # Set adata var_names to be new ec index
    adata.var_names = np.array(np.arange(adata.shape[1]), dtype=str) # var names are set names
    # Create dictionary mapping tx indices to tx ids for each ec
    ec2tx = {}
    for (ec, txs) in ecdict.items():
        ec2tx[ec] = list(t2enst.loc[txs].enst.values)
    # Create mapping of each ec to the genes whose transcripts are contained in it
    ec2g = {}
    key_errors = {}
    ec_error = []
    for (ec, txs) in ec2tx.items():
        try:
            # get unique gene names for each ec list of transcripts
            ec2g[ec] = list(np.unique(t2g.loc[txs].gname.values))
            ec_error.append(False)
        except KeyError:
            # if there's a tx that doesn't have a t2g mapping
            valid_txs = [i for i in txs if i in t2g.index]
            if len(valid_txs) > 0:
                # if there is at least one valid gene for the ec txs use them instead
                ec2g[ec] = list(np.unique(t2g.loc[valid_txs].gname.values))
            else:
                # Otherwise, fill in none for ec2g mapping
                ec2g[ec] = "None"
            ec_error.append(True)
            # If the ec list of transcripts contains a tx not in the tx to gene mapping dict, note the ec and txs that
            # are the problem
            errors = [i for i in txs if i not in t2g.index]
            key_errors[ec] = errors
    adata.var["ec_tx_error"] = ec_error
    # Create an inverse mapping dictionary getting all ecs and their txs for a given gene
    inv_map = {}
    for ec, genes in ec2g.items():
        for gene in genes:
            # Get the current dict of ec's for that gene
            inv_map[gene] = inv_map.get(gene, {})
            # Append the current ec
            inv_map[gene][ec] = ec2tx[ec]
    # save the dicts
    logger.info(f"Number of ec errors: {len(key_errors)}")
    adata.uns['key_errors'] = key_errors
    return ec2tx, ec2g, inv_map




"""
Filters cells based on gene number, total counts, and % mitochondrial
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
from anndata import AnnData
from typing import Tuple

# scrnatools package imports
from .._configs import configs
from .._utils import debug

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


@debug(logger, configs)
def filter_cells(
        adata: AnnData,
        genes_thresholds: Tuple[int, int],
        counts_thresholds: Tuple[int, int],
        mt_threshold: int = 10,
) -> AnnData:
    """
    Filters cells based on gene number, total counts, and % mitochondrial
    Parameters
    ----------
    adata
        The AnnData with the data to filter
    genes_thresholds
        A Tuple of thresholds for the number of genes per cell with 'gene_thresholds[0]' being the lower bound and
        'gene_thresholds[1]' being the upper bound (both exclusive).
    counts_thresholds
        A Tuple of thresholds for the number of total counts cell with 'count_thresholds[0]' being the lower bound and
        'count_thresholds[1]' being the upper bound (both exclusive).
    mt_threshold
        The maximum percent mitochondrial reads per cell. Default 10%.
    Returns
    -------
    An AnnData object with cells that don't pass the thresholds filtered out
    """
    logger.info(f"Number of cells before QC filtering: {len(adata.obs)}")
    filtered_adata = adata[adata.obs.pct_counts_mt < mt_threshold].copy()
    filtered_adata = filtered_adata[filtered_adata.obs.total_counts < counts_thresholds[1]]
    filtered_adata = filtered_adata[filtered_adata.obs.total_counts > counts_thresholds[0]]
    filtered_adata = filtered_adata[filtered_adata.obs.n_genes_by_counts < genes_thresholds[1]]
    filtered_adata = filtered_adata[filtered_adata.obs.n_genes_by_counts > genes_thresholds[0]].copy()
    logger.info(f"Number of cells after QC filtering: {len(filtered_adata.obs)}")
    return filtered_adata



"""
scrnatools package init

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

from ._configs import configs
from . import plotting as pl
from . import qc
from . import tools as tl
